import ast
import copy
import os
import shutil
import sys
import tempfile
import types
from functools import wraps
from importlib import import_module

import yaml

from vlab.io.io import load, save

BASE_KEY = "_base_"

_CONFIG = None


def inject_config(cfg):
    """
    Injects the config into the global config.

    See `configurable` for example usage.
    """
    global _CONFIG
    _CONFIG = cfg


def configurable(fn_or_cls):
    """
    A simple decorator that injects the global config into the function or class.

    Example:
        >>> @configurable
        ... def my_func(a, b):
        ...     return a + b
        ...
        >>> inject_config(Config({"my_func": {"a": 1, "b": 2}}))
        >>> my_func()
        3
        >>> @configurable
        ... class MyClass:
        ...     def __init__(self, a, b):
        ...         self.a = a
        ...         self.b = b
        ...     def my_func(self):
        ...         return self.a + self.b
        ...
        >>> inject_config(Config({"MyClass": {"a": 1, "b": 2}}))
        >>> MyClass().my_func()
        3
    """

    if isinstance(fn_or_cls, types.FunctionType):

        @wraps(fn_or_cls)
        def wrapper(*args, **kwargs):
            if _CONFIG is None:
                raise RuntimeError("Config not injected.")
            fn_or_cls_cfg = _CONFIG[fn_or_cls.__name__]
            if isinstance(fn_or_cls_cfg, ConfigDict):
                fn_or_cls_cfg = fn_or_cls_cfg.to_dict()

            return fn_or_cls(*args, **kwargs, **fn_or_cls_cfg)

        return wrapper
    elif isinstance(fn_or_cls, type):

        class Wrapper(fn_or_cls):
            def __init__(self, *args, **kwargs):
                if _CONFIG is None:
                    raise RuntimeError("Config not injected.")
                fn_or_cls_cfg = _CONFIG[fn_or_cls.__name__]
                if isinstance(fn_or_cls_cfg, ConfigDict):
                    fn_or_cls_cfg = fn_or_cls_cfg.to_dict()
                super().__init__(*args, **kwargs, **fn_or_cls_cfg)

            def __repr__(self):
                return "{}({})".format(fn_or_cls.__name__, self.__dict__)

            def __str__(self):
                return self.__repr__()

        return Wrapper
    else:
        raise TypeError("@configurable can only be applied to functions or classes")


class ConfigDict(dict):
    def __init__(self, d=None, **kwargs):
        super().__init__()
        d = {} if d is None else d
        if kwargs:
            d.update(kwargs)
        for k, v in d.items():
            self[k] = v

    def __setitem__(self, name, value):
        def is_dict(x):
            return isinstance(x, dict) and not isinstance(x, ConfigDict)

        if is_dict(value):
            value = ConfigDict(value)
        elif isinstance(value, list) or isinstance(value, tuple):
            value = type(value)(ConfigDict(v) if is_dict(v) else v for v in value)

        super().__setitem__(name, value)
        super().__setattr__(name, value)

    __setattr__ = __setitem__

    def update(self, d=None, **kwargs):
        d = {} if d is None else d
        if kwargs:
            d.update(kwargs)
        for k, v in d.items():
            self[k] = v

    def pop(self, k, d=None):
        delattr(self, k)
        return super().pop(k, d)

    def copy(self):
        return copy.copy(self)

    def to_dict(self):
        return {k: v.to_dict() if isinstance(v, ConfigDict) else v for k, v in self.items()}


class Config:
    """
    A facility for configuration.

    It supports yaml/json/py formats. It allows to access values as a dict or as attributes.

    Example:
        >>> cfg = Config({"a": 1, "b": {"c": 2}})
        >>> cfg.a
        1
        >>> cfg.b.c
        2
        >>> cfg["a"]
        1
        >>> cfg = Config.from_file("config.yaml")
        >>> cfg.key1
        "value1"
    """

    @staticmethod
    def from_file(path):
        return Config(cfg_dict=Config._file_to_dict(path))

    @staticmethod
    def _merge_a_into_b(a, b):
        """
        Values in a override values in b.
        """
        b = b.copy()
        for k, v in a.items():
            if isinstance(v, dict):
                if k in b:
                    b[k] = Config._merge_a_into_b(v, b[k])
                else:
                    b[k] = v
            else:
                b[k] = v
        return b

    @staticmethod
    def _file_to_dict(path):
        """
        Load config from a file.
        """
        path = os.path.abspath(os.path.expanduser(path))
        file_ext = path.split(".")[-1].lower()
        if file_ext not in ["yaml", "yml", "json", "py"]:
            raise ValueError(f"Unsupported file extension: {file_ext}")

        if file_ext == "py":
            cfg_dict = Config._python_file_to_dict(path)
        else:
            cfg_dict = load(path)

        if BASE_KEY in cfg_dict:
            cfg_dir = os.path.dirname(path)
            base_filename = cfg_dict.pop(BASE_KEY)
            base_cfg_dict = Config._file_to_dict(os.path.join(cfg_dir, base_filename))
            cfg_dict = Config._merge_a_into_b(cfg_dict, base_cfg_dict)

        return cfg_dict

    @staticmethod
    def _python_file_to_dict(path):
        with open(path, encoding="utf-8") as f:
            content = f.read()
        try:
            ast.parse(content)
        except SyntaxError as e:
            raise SyntaxError(f"There are syntax errors in config file {path}: {e}")

        with tempfile.TemporaryDirectory() as temp_cfg_dir:
            temp_cfg_file = tempfile.NamedTemporaryFile(dir=temp_cfg_dir, suffix=".py")
            temp_cfg_name = os.path.basename(temp_cfg_file.name)
            shutil.copyfile(path, temp_cfg_file.name)
            temp_module_name = os.path.splitext(temp_cfg_name)[0]

            sys.path.insert(0, temp_cfg_dir)
            module = import_module(temp_module_name)
            sys.path.pop(0)
            cfg_dict = {}
            for name, value in module.__dict__.items():
                if not (
                    name.startswith("__")
                    or isinstance(value, types.ModuleType)
                    or isinstance(value, types.FunctionType)
                ):
                    cfg_dict[name] = value
            del sys.modules[temp_module_name]  # delete imported module
            temp_cfg_file.close()
        return cfg_dict

    def __init__(self, cfg_dict=None, cfg_path=None):
        """
        Initialize a config object.

        Args:
            cfg_dict (dict): A dict to initialize the config object.
            cfg_path (str): A path to a file to initialize the config object.
        """
        if cfg_dict is not None:
            if not isinstance(cfg_dict, dict):
                raise TypeError("cfg_dict must be a dict")
            super().__setattr__("_cfg_dict", ConfigDict(cfg_dict))
        elif cfg_path is not None:
            super().__setattr__("_cfg_dict", ConfigDict(Config._file_to_dict(cfg_path)))
        else:
            raise ValueError("cfg_dict or cfg_path must be provided")

        super().__setattr__("_freeze", False)

    def __repr__(self):
        _cfg_dict = super().__getattribute__("_cfg_dict")
        return f"Config: {_cfg_dict.__repr__()}"

    def __len__(self):
        return len(super().__getattribute__("_cfg_dict"))

    def __str__(self):
        return super().__getattribute__("_cfg_dict").__str__()

    def __getattr__(self, name):
        return getattr(super().__getattribute__("_cfg_dict"), name)

    def __getitem__(self, name):
        return super().__getattribute__("_cfg_dict").__getitem__(name)

    def __setattr__(self, name, value):
        if super().__getattribute__("_freeze"):
            raise AttributeError("Config is frozen")
        super().__getattribute__("_cfg_dict").__setattr__(name, value)

    def __setitem__(self, name, value):
        if super().__getattribute__("_freeze"):
            raise AttributeError("Config is frozen")
        super().__getattribute__("_cfg_dict").__setitem__(name, value)

    def __iter__(self):
        return iter(super().__getattribute__("_cfg_dict"))

    def __copy__(self):
        cls = self.__class__
        other = cls.__new__(cls)
        other.__dict__.update(self.__dict__)
        return other

    def __deepcopy__(self, memo):
        cls = self.__class__
        other = cls.__new__(cls)
        memo[id(self)] = other
        for key, value in self.__dict__.items():
            super(Config, other).__setattr__(key, copy.deepcopy(value, memo))
        return other

    def merge_from_list(self, other_cfg_list):
        """
        Values in other_cfg_list override values in self.
        """
        if not len(other_cfg_list) % 2 == 0:
            raise ValueError("cfg_list must be a list of key-value pairs")
        other_cfg_dict = {k: v for k, v in zip(other_cfg_list[::2], other_cfg_list[1::2])}
        cfg_dict = super().__getattribute__("_cfg_dict")
        super().__setattr__("_cfg_dict", Config._merge_a_into_b(other_cfg_dict, cfg_dict))

    def merge_from_dict(self, other_cfg_dict):
        """
        Values in other_cfg_dict override values in self.
        """
        cfg_dict = super().__getattribute__("_cfg_dict")
        super().__setattr__("_cfg_dict", Config._merge_a_into_b(other_cfg_dict, cfg_dict))

    def to_file(self, output_path):
        return save(super().__getattribute__("_cfg_dict").to_dict(), output_path)

    def freeze(self):
        """
        Freeze the config object.
        """
        super().__setattr__("_freeze", True)

    def unfreeze(self):
        """
        Unfreeze the config object.
        """
        super().__setattr__("_freeze", False)

    def yaml_text(self):
        return yaml.dump(super().__getattribute__("_cfg_dict").to_dict())
