import os
import socket

import torch
from torch import distributed as dist


def default_dist_setup(rank, world_size, port=12345):
    """
    Setup torch distributed training.

    Example:
        >>> from torch import multiprocessing
        >>> import vlab
        >>> def run(rank, world_size):
        >>>     vlab.default_dist_setup(rank, world_size)
        >>>     # ... training loop
        >>>     vlab.default_dist_cleanup()
        >>> if __name__ == "__main__":
        >>>     n_gpus = torch.cuda.device_count()
        >>>     multiprocessing.spawn(run, nprocs=n_gpus, args=(world_size,))
    """

    if not is_port_available(int(port)):
        port = find_free_port()

    os.environ["MASTER_ADDR"] = "localhost"
    os.environ["MASTER_PORT"] = str(port)

    dist.init_process_group("nccl", rank=rank, world_size=world_size)
    torch.cuda.set_device(rank)


def default_dist_cleanup():
    dist.destroy_process_group()


def find_free_port() -> int:
    with socket.socket() as sock:
        sock.bind(("", 0))  # bind to a free port provided by the host.
        return sock.getsockname()[1]  # return the port number assigned.


def is_port_available(port: int, host="localhost"):
    with socket.socket() as sock:
        if sock.connect_ex((host, port)) == 0:
            return False
        else:
            return True


if __name__ == "__main__":
    print(is_port_available(3000))
    print(find_free_port())
