import logging
from datetime import datetime

from rich.highlighter import ReprHighlighter
from rich.table import Table
from rich.text import Text

from vlab.common.console import _get_console

_ROOT_LOGGER = logging.Logger("vlab")


def set_logger(log_level="INFO", log_file=None, file_mode="w"):
    _ROOT_LOGGER.setLevel(log_level)
    _ROOT_LOGGER.propagate = False

    if not any([isinstance(x, RichHandler) for x in _ROOT_LOGGER.handlers]):
        console = _get_console()
        formatter = logging.Formatter("%(message)s")
        rich_handler = RichHandler(log_level, console)
        rich_handler.setFormatter(formatter)
        _ROOT_LOGGER.addHandler(rich_handler)

    if log_file is not None:
        # overwrite existing file handler
        for hdlr in _ROOT_LOGGER.handlers:
            if isinstance(hdlr, logging.FileHandler):
                _ROOT_LOGGER.removeHandler(hdlr)
        file_handler = logging.FileHandler(log_file, file_mode)
        formatter = logging.Formatter("[%(asctime)s] %(levelname)s %(message)s", datefmt="%H:%M:%S")
        file_handler.setFormatter(formatter)
        file_handler.setLevel(log_level)
        _ROOT_LOGGER.addHandler(file_handler)


class RichHandler(logging.Handler):
    def __init__(self, level, console, datefmt="%H:%M:%S"):
        super().__init__(level=level)
        self.console = console
        self.datefmt = datefmt
        self.highlighter = ReprHighlighter()

    def emit(self, record):
        message = self.format(record)
        message_text = Text(message)
        message_text = self.highlighter(message_text)
        log_renderable = self.render(record, message_text)

        try:
            self.console.print(log_renderable)
        except Exception:
            self.handleError(record)

    def render(self, record, message_text):
        level_no = record.levelno
        level_name = record.levelname

        output = Table.grid(padding=(0, 1))
        output.expand = True

        output.add_column()
        output.add_column(justify="left", overflow="fold", ratio=1)

        row = []
        log_time = datetime.fromtimestamp(record.created)
        text = Text()
        text.append("[", style="log.time")
        text.append(log_time.strftime(self.datefmt), style="log.time")
        text.append("]", style="log.time")

        if (level_no == logging.WARNING) or (level_no == logging.ERROR):
            text.append(" ")
            level_text = Text.styled(level_name, f"logging.level.{level_name.lower()}")
            text.append(level_text)

        row.append(text)
        row.append(message_text)

        output.add_row(*row)

        return output


def debug(*args, **kwargs):
    if len(_ROOT_LOGGER.handlers) == 0:
        set_logger()
    _ROOT_LOGGER.debug(*args, **kwargs)


def info(*args, **kwargs):
    if len(_ROOT_LOGGER.handlers) == 0:
        set_logger()
    _ROOT_LOGGER.info(*args, **kwargs)


def warning(*args, **kwargs):
    if len(_ROOT_LOGGER.handlers) == 0:
        set_logger()
    _ROOT_LOGGER.warning(*args, **kwargs)


def error(*args, **kwargs):
    if len(_ROOT_LOGGER.handlers) == 0:
        set_logger()
    _ROOT_LOGGER.error(*args, **kwargs)


def critical(*args, **kwargs):
    if len(_ROOT_LOGGER.handlers) == 0:
        set_logger()
    _ROOT_LOGGER.critical(*args, **kwargs)


def rule(msg):
    console = _get_console()
    console.rule(msg)


if __name__ == "__main__":
    set_logger(log_level="DEBUG")
    info("here is a info")
    info({"key": "value"})
    debug("here is a debug")
    info("highlight number: 0.123")
    warning("here is a warning")
    error("here is a error")
    info("here is a very looooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooong log")
    rule("begin something")
