import numpy as np
import torch
import torch.linalg
import torch.nn.functional as F

from vlab.core import Backends, get_backend


def matmul(x1, x2):
    backend = get_backend(x1)
    if backend == Backends.NUMPY:
        return np.matmul(x1, x2)
    elif backend == Backends.TORCH:
        return torch.matmul(x1, x2)
    else:
        raise TypeError(f"Unsupported backend {backend}")


def cross(x1, x2, dim=None):
    backend = get_backend(x1)
    if backend == Backends.NUMPY:
        return np.cross(x1, x2, axis=dim)
    elif backend == Backends.TORCH:
        return torch.cross(x1, x2, dim=dim)
    else:
        raise TypeError(f"Unsupported backend {backend}")


def norm(x, p=2, dim=None, keepdim=False):
    backend = get_backend(x)
    if backend == Backends.NUMPY:
        return np.linalg.norm(x, ord=p, axis=dim, keepdims=keepdim)
    elif backend == Backends.TORCH:
        return torch.norm(x, p=p, dim=dim, keepdim=keepdim)
    else:
        raise TypeError(f"Unsupported backend {backend}")


def normalize(x, p=2, dim=None, eps=1e-12):
    backend = get_backend(x)
    if backend == Backends.NUMPY:
        x_norm = np.clip(np.linalg.norm(x, ord=p, axis=dim, keepdims=True), a_min=eps, a_max=None)
        return x / x_norm
    elif backend == Backends.TORCH:
        return F.normalize(x, p=p, dim=dim, eps=eps)
    else:
        raise TypeError(f"Unsupported backend {backend}")
