import cv2

CV2_INTERP_CODES = {
    "nearest": cv2.INTER_NEAREST,
    "bilinear": cv2.INTER_LINEAR,
    "bicubic": cv2.INTER_CUBIC,
    "area": cv2.INTER_AREA,
    "lanczos": cv2.INTER_LANCZOS4,
}


def imresize(img, size, interpolation="bilinear"):
    """
    Resize image to a given size.

    Args:
        img (ndarray): The input image.
        size (tuple[int]): Target size (w, h).
        interpolation (str): Interpolation method, accepted values are
            "nearest", "bilinear", "bicubic", "area", "lanczos"

    Returns:
        ndarray: The resized image.
    """
    resized_img = cv2.resize(img, size, interpolation=CV2_INTERP_CODES[interpolation])
    return resized_img


def imrescale(img, scale, interpolation="bilinear"):
    """Resize image while keeping the aspect ratio.

    Args:
        img (ndarray): The input image.
        scale (float): The scaling factor or maximum size.
        interpolation (str): Same as `resize`.

    Returns:
        ndarray: The rescaled image.
    """
    h, w = img.shape[:2]

    if isinstance(scale, (float, int)):
        if scale <= 0:
            raise ValueError(f"Invalid scale {scale}, must be positive.")
    else:
        raise TypeError(f"Scale must be a number, but got {type(scale)}")

    if scale == 1.0:
        return img

    # int(num + 0.5) is the same as round(num) in Python 3.
    new_h, new_w = int(h * float(scale) + 0.5), int(w * float(scale) + 0.5)
    rescaled_img = imresize(img, (new_w, new_h), interpolation=interpolation)
    return rescaled_img
