from typing import Callable, Sequence, Union

import torch
import torch.nn as nn
import torch.nn.functional as F


class MLPv1(nn.Module):
    def __init__(self, layer_sizes: Sequence[int], actv_func: Union[Callable, str] = "ReLU", last_actv: bool = False):
        """
        Args:
            layer_sizes: sequence of layer sizes
            actv_func: callable func or str, accepted strs are "ReLU", "LeakyReLU", "Tanh", "Sigmoid", etc,
                i.e. attributes of torch.nn.
            last_actv: whether to use activation function after last layer
        """
        super().__init__()

        if isinstance(actv_func, str):
            try:
                actv_func = getattr(nn, actv_func)
            except AttributeError:
                raise ValueError(f"Unknown actv_func: {actv_func}. Accepted values are: ReLU, Sigmoid, Tanh, etc.")

        layers = []
        for i, (in_size, out_size) in enumerate(zip(layer_sizes[:-1], layer_sizes[1:])):
            layers.append(nn.Linear(in_size, out_size))
            if last_actv or i < len(layer_sizes) - 2:
                layers.append(actv_func())
        self.layers = nn.Sequential(*layers)

    def forward(self, x):
        return self.layers(x)


class MLPv2(nn.Module):
    def __init__(
        self,
        depth: int,
        width: int,
        input_dim: int,
        output_dim: int = -1,
        hidden_actv: Union[Callable, str] = "relu",
        output_actv: Union[Callable, str] = "identity",
        skip_layers: Sequence[int] = tuple(),
    ):
        """
        Args:
            depth: number of layers
            width: width of each layer
            input_dim: input dimension
            output_dim: output dimension, if -1, output_dim = width
            hidden_actv, output_actv: callable func or str, accepted strs are "relu", "leaky_relu", "tanh", "sigmoid",
                etc, i.e. attributes of torch.nn.functional.
            skip_layers: sequence of layers with skip connections
        """
        super().__init__()
        if isinstance(hidden_actv, str):
            try:
                hidden_actv = getattr(F, hidden_actv)
            except AttributeError:
                raise ValueError(f"Unknown activation function: {hidden_actv}.")
        if isinstance(output_actv, str):
            if output_actv == "identity":
                output_actv = lambda x: x  # noqa: E731
            else:
                try:
                    output_actv = getattr(F, output_actv)
                except AttributeError:
                    raise ValueError(f"Unknown activation function: {output_actv}.")

        output_dim = output_dim if output_dim > 0 else width

        layers = []
        for i in range(depth):
            if i == 0:
                layer_in_dim = input_dim
            elif i in skip_layers:
                layer_in_dim = width + input_dim
            else:
                layer_in_dim = width
            layer_out_dim = width if i < depth - 1 else output_dim
            layers.append(nn.Linear(layer_in_dim, layer_out_dim))

        self.layers = nn.ModuleList(layers)
        self.hidden_actv = hidden_actv
        self.output_actv = output_actv
        self.depth = depth
        self.skip_layers = skip_layers

    def forward(self, x):
        ori_x = x
        for i in range(self.depth):
            if i in self.skip_layers:
                x = torch.cat((x, ori_x), dim=-1)
            x = self.layers[i](x)
            if i < self.depth - 1:
                x = self.hidden_actv(x)
        return self.output_actv(x)


MLP = MLPv1

if __name__ == "__main__":
    # model = MLP([256, 512, 256, 128], actv_func="ReLU")
    # print(model)
    # rand_x = torch.rand(4, 256)
    # output = model(rand_x)
    # assert output.shape == (4, 128)

    model = MLPv2(depth=8, width=128, input_dim=3, output_dim=1, skip_layers=(4,))
    print(model)
    rand_x = torch.rand(4, 3)
    output = model(rand_x)
    assert output.shape == (4, 1)
