import vlab


def so3_exp_map(log_rot, eps=0.0001):
    """
    Convert a batch of logarithmic representations of rotation matrices `log_rot`
    to a batch of 3x3 rotation matrices using Rodrigues formula [1].

    In the logarithmic representation, each rotation matrix is represented as
    a 3-dimensional vector (`log_rot`) who's l2-norm and direction correspond
    to the magnitude of the rotation angle and the axis of rotation respectively.
    The conversion has a singularity around `log(R) = 0`
    which is handled by clamping controlled with the `eps` argument.

    Args:
        log_rot: Batch of vectors of shape [..., 3]
        eps: A float constant handling the conversion singularity.
    Returns:
        Batch of rotation matrices of shape [..., 3, 3]
    Raises:
        ValueError if `log_rot` is of incorrect shape.

    [1] https://en.wikipedia.org/wiki/Rodrigues%27_rotation_formula
    """
    return _so3_exp_map(log_rot, eps=eps)[0]


def _so3_exp_map(log_rot, eps=0.0001):
    if log_rot.shape[-1] != 3:
        raise ValueError("Input tensor shape has to be [..., 3].")

    norms = vlab.sum(log_rot * log_rot, dim=-1)
    # phis ... rotation angles
    rot_angles = vlab.sqrt(vlab.clamp(norms, eps))

    rot_angles_inv = 1.0 / rot_angles
    fac1 = rot_angles_inv * vlab.sin(rot_angles)
    fac2 = rot_angles_inv * rot_angles_inv * (1.0 - vlab.cos(rot_angles))
    skews = hat(log_rot)
    skews_square = vlab.matmul(skews, skews)

    R = fac1[..., None, None] * skews + fac2[..., None, None] * skews_square + vlab.eye(3, ref=log_rot)[None]

    return R, rot_angles, skews, skews_square


def hat(v):
    """
    Compute the Hat operator [1] of a batch of 3D vectors.
    Args:
        v: Batch of vectors of shape [..., 3]
    Returns:
        Batch of skew-symmetric matrices of shape [..., 3, 3] where each matrix is of the form:
            `[    0  -v_z   v_y ]
             [  v_z     0  -v_x ]
             [ -v_y   v_x     0 ]`
    Raises:
        ValueError if `v` is of incorrect shape.
    [1] https://en.wikipedia.org/wiki/Hat_operator
    """

    if v.shape[-1] != 3:
        raise ValueError("Input vectors have to be 3-dimensional.")

    h = vlab.zeros((*v.shape[:-1], 3, 3), ref=v)
    x, y, z = vlab.unbind(v, dim=-1)

    h[..., 0, 1] = -z
    h[..., 0, 2] = y
    h[..., 1, 0] = z
    h[..., 1, 2] = -x
    h[..., 2, 0] = -y
    h[..., 2, 1] = x

    return h
