import vlab


def transform(pts, tf_mat):
    """
    Apply a transformation matrix on a set of 3D points.

    Args:
        pts: 3D points, could be [..., N, 3]
        tf_mat: Transformation matrix, could be [..., 4, 4]

        The dimension number of pts and tf_mat should be the same.

    Returns:
        Transformed pts.
    """
    if pts.ndim != tf_mat.ndim:
        raise ValueError("The dimension number of pts and tf_mat should be the same.")

    homo_pts = to_homo(pts)
    # `homo_pts @ tf_mat.T` or `(tf_mat @ homo_pts.T).T`
    new_pts = vlab.matmul(homo_pts, vlab.transpose(tf_mat, -2, -1))
    return new_pts[..., :3]


def rotate(pts, rot_mat):
    """
    Apply a rotation matrix on a set of 3D points.

    Args:
        pts: 3D points, could be [..., N, 3]
        rot_mat: Rotation matrix, could be [..., 3, 3]

        The dimension number of pts and rot_mat should be the same.

    Returns:
        Rotated pts.
    """
    if pts.ndim != rot_mat.ndim:
        raise ValueError("The dimension number of pts and rot_mat should be the same")

    # `pts @ rot_mat.T` or `(rot_mat @ pts.T).T`
    new_pts = vlab.matmul(pts, vlab.transpose(rot_mat, -2, -1))
    return new_pts


def project(pts, intr_mat, return_depth=False):
    """
    Project 3D points in the camera space to the image plane.

    Args:
        pts: 3D points, could be Nx3 or BxNx3.
        intr_mat: Intrinsic matrix, could be 3x3 or Bx3x3.

    Returns:
        pixels: the order is uv other than xy.
        depth (if return_depth): depth in the camera space.
    """
    pts = vlab.clone(pts)
    new_pts = pts / pts[..., 2:3]
    new_pts = vlab.matmul(new_pts, vlab.transpose(intr_mat, -2, -1))

    if not return_depth:
        return new_pts[..., :2]
    else:
        return new_pts[..., :2], pts[..., 2]


def unproject(pixels, depth, intr_mat):
    """
    Unproject pixels in the image plane to 3D points in the camera space.

    Args:
        pixels: Pixels in the image plane, could be Nx2 or BxNx2. The order is uv rather than xy.
        depth: Depth in the camera space, could be N, Nx1, BxN or BxNx1.
        intr_mat: Intrinsic matrix, could be 3x3 or Bx3x3.
    Returns:
        pts: 3D points, Nx3 or BxNx3.
    """
    if depth.ndim < pixels.ndim:
        depth = depth[..., None]  # N -> Nx1, BxN -> BxNx1
    principal_point = vlab.unsqueeze(intr_mat[..., :2, 2], dim=-2)  # 1x2, Bx1x2
    focal_length = vlab.cat([intr_mat[..., 0:1, 0:1], intr_mat[..., 1:2, 1:2]], dim=-1)  # 1x2, Bx1x2
    xys = (pixels - principal_point) * depth / focal_length
    pts = vlab.cat([xys, depth], dim=-1)
    return pts


# Ref: https://math.stackexchange.com/a/1315407/757569
def inverse(rot_or_tf_mat):
    if rot_or_tf_mat.shape[-1] == 3:  # rotation matrix
        new_mat = vlab.transpose(rot_or_tf_mat, -2, -1)
    else:  # transformation matrix
        new_rot_mat = vlab.transpose(rot_or_tf_mat[..., :3, :3], -2, -1)
        ori_tl = vlab.unsqueeze(rot_or_tf_mat[..., :3, 3], dim=-2)  # 1x3, Bx1x3
        new_tl = vlab.squeeze(-rotate(ori_tl, new_rot_mat), dim=-2)  # 3, Bx3
        new_mat = rot_tl_to_tf_mat(new_rot_mat, new_tl)
    return new_mat


# Ref: https://www.scratchapixel.com/lessons/mathematics-physics-for-computer-graphics/geometry/row-major-vs-column-major-vector # noqa
def swap_major(rot_or_tf_mat):
    return vlab.transpose(rot_or_tf_mat, -2, -1)


def rot_tl_to_tf_mat(rot_mat, tl=None):
    """
    Build transformation matrix with rotation matrix and translation vector.

    Args:
        rot_matx: rotation matrix, could be ...x3x3.
        tl: translation vector, could be ...x3. If None, translation will be 0.
    Returns:
        tf_mat, transformation matrix.
    """

    tf_mat = vlab.repeat(vlab.eye(4, ref=rot_mat), rot_mat.shape[:-2] + (1, 1))
    tf_mat[..., :3, :3] = rot_mat
    if tl is not None:
        tf_mat[..., :3, 3] = tl
    return tf_mat


def to_homo(pts_3d):
    """
    Convert Cartesian 3D points to Homogeneous 4D points.

    Args:
      pts_3d: 3D points in Cartesian coord, could be ...x3.
    Returns:
      ...x4 points in the Homogeneous coord.
    """
    return vlab.cat([pts_3d, vlab.ones_like(pts_3d[..., :1])], dim=-1)


def from_homo(pts_homo):
    return pts_homo[..., :3] / pts_homo[..., -1:]
