import multiprocessing
from typing import Sized

from rich.progress import (
    BarColumn,
    Progress,
    ProgressColumn,
    TextColumn,
    TimeRemainingColumn,
)
from rich.text import Text

from vlab.common.console import _get_console


class ProgressBar(Progress):
    def __init__(self):
        console = _get_console()
        super().__init__(
            TextColumn("[progress.description]{task.description}"),
            BarColumn(),
            TextColumn("[magenta]{task.completed}/{task.total}"),
            TimeElapsedColumn(compact=True),
            TimeRemainingColumn(compact=True),
            console=console,
        )

    def step(self, task_id, description=None):
        if description is not None:
            self.update(task_id, advance=1, description=description)
        else:
            self.update(task_id, advance=1)


class TimeElapsedColumn(ProgressColumn):
    def __init__(self, compact=True):
        self.compact = compact
        super().__init__()

    def render(self, task):
        elapsed = task.finished_time if task.finished else task.elapsed
        style = "progress.elapsed"

        if elapsed is None:
            return Text("--:--" if self.compact else "-:--:--", style=style)

        minutes, seconds = divmod(int(elapsed), 60)
        hours, minutes = divmod(minutes, 60)

        if self.compact and not hours:
            formatted = f"{minutes:02d}:{seconds:02d}"
        else:
            formatted = f"{hours:d}:{minutes:02d}:{seconds:02d}"

        return Text(formatted, style=style)


def track(sequence, total=None, description="", disable=False):
    if not disable:
        pbar = ProgressBar()
        with pbar:
            yield from pbar.track(sequence, total=total, description=description)
    else:
        yield from sequence


def parallel_track(func, tasks, process_num, total=None, description="", disable=False, keep_order=True, chunk_size=1):
    if total is None:
        if isinstance(tasks, Sized):
            task_total = float(len(tasks))
        else:
            raise ValueError(f"unable to get size of {tasks!r}, please specify 'total'")
    else:
        task_total = total

    pool = multiprocessing.Pool(processes=process_num)

    if keep_order:
        gen = pool.imap(func, tasks, chunk_size)
    else:
        gen = pool.imap_unordered(func, tasks, chunk_size)

    results = []
    for result in track(gen, total=task_total, description=description, disable=disable):
        results.append(result)

    pool.close()
    pool.join()
    return results


def parallel_map(func, tasks, process_num, keep_order=True, chunk_size=1):
    pool = multiprocessing.Pool(processes=process_num)

    if keep_order:
        gen = pool.imap(func, tasks, chunk_size)
    else:
        gen = pool.imap_unordered(func, tasks, chunk_size)

    results = []
    for result in gen:
        results.append(result)

    pool.close()
    pool.join()
    return results


if __name__ == "__main__":
    import time

    import vlab

    for i in track(range(10)):
        vlab.info(str(i))
        time.sleep(0.2)
