import os
import shutil
import subprocess

import cv2

from vlab.image_proc.io import imread, imwrite
from vlab.utils.misc import requires_executable


class VideoReader:
    """
    Video class with similar usage to a list object.
    """

    def __init__(self, filepath, to_rgb=True):
        self.filepath = filepath
        self.cap = cv2.VideoCapture(filepath)
        self.frame_count = int(self.cap.get(cv2.CAP_PROP_FRAME_COUNT))
        self.fps = self.cap.get(cv2.CAP_PROP_FPS)
        self._width = self.cap.get(cv2.CAP_PROP_FRAME_WIDTH)
        self._height = self.cap.get(cv2.CAP_PROP_FRAME_HEIGHT)
        self.to_rgb = to_rgb

    def __len__(self):
        return self.frame_count

    @property
    def shape(self):
        return (self.frame_count, self.height, self.width, 3)

    @property
    def height(self):
        return self._height

    @property
    def width(self):
        return self._width

    def write_images(self, img_dir, auto_mkdirs=False, fmt_str="{:06d}.png"):
        if auto_mkdirs:
            os.makedirs(img_dir, exist_ok=True)
        for i, img in enumerate(self):
            imwrite(img, os.path.join(img_dir, fmt_str.format(i)))

    def __getitem__(self, index):
        self.cap.set(cv2.CAP_PROP_POS_FRAMES, index)
        ret, frame = self.cap.read()
        if not ret:
            raise IndexError("Video index out of range")
        if self.to_rgb:
            if frame.shape[-1] == 3:
                frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
            elif frame.shape[-1] == 4:
                frame = cv2.cvtColor(frame, cv2.COLOR_BGRA2RGBA)
        return frame

    def __iter__(self):
        for i in range(self.frame_count):
            yield self[i]

    def __repr__(self):
        return f"VideoReader({self.filepath})"

    def __str__(self):
        return f"VideoReader({self.filepath})"

    def __del__(self):
        self.cap.release()


default_codecs = {
    "avi": "MJPG",
    "mp4": "mp4v",
    "mov": "XVID",
}


def images2video_cv2(images_or_dir, filepath, fps=30, fourcc=None, from_rgb=True, auto_mkdirs=False):
    if auto_mkdirs:
        os.makedirs(os.path.dirname(filepath), exist_ok=True)

    if fourcc is None:  # get fourcc from file extension
        ext = os.path.splitext(filepath)[1][1:]
        fourcc = default_codecs.get(ext, "XVID")

    if isinstance(images_or_dir, str):
        images = [imread(os.path.join(images_or_dir, f)) for f in sorted(os.listdir(images_or_dir))]
    else:
        images = images_or_dir

    height, width = images[0].shape[:2]
    writer = cv2.VideoWriter(filepath, cv2.VideoWriter_fourcc(*fourcc), fps, (width, height))
    for image in images:
        if from_rgb:
            if image.shape[-1] == 3:
                image = cv2.cvtColor(image, cv2.COLOR_RGB2BGR)
            elif image.shape[-1] == 4:
                image = cv2.cvtColor(image, cv2.COLOR_RGBA2BGRA)
        writer.write(image)
    writer.release()


@requires_executable("ffmpeg")
def images2video(images_or_dir, output_path, fps=30, from_rgb=True, auto_mkdirs=False, image_ext="png"):
    cmd_fmt = "ffmpeg -framerate {} -pattern_type glob -i '{}' -c:v libx264 -pix_fmt yuv420p {} -y"

    if auto_mkdirs:
        os.makedirs(os.path.dirname(output_path), exist_ok=True)

    if not isinstance(images_or_dir, str):
        images_dir = os.path.join(os.path.dirname(output_path), "images4video")
        os.makedirs(images_dir, exist_ok=True)
        for i, img in enumerate(images_or_dir):
            imwrite(img, os.path.join(images_dir, f"{i:06d}.{image_ext}"), from_rgb=from_rgb)
    else:
        images_dir = images_or_dir

    full_glob = os.path.join(images_dir, "*.{}".format(image_ext))
    full_cmd = cmd_fmt.format(fps, full_glob, output_path)

    ret_code = subprocess.call(full_cmd, shell=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)

    if ret_code != 0:
        raise RuntimeError("ffmpeg failed")
    else:
        shutil.rmtree(images_dir, ignore_errors=True)
        return ret_code
