import numpy as np


def _make_color_wheel():
    """
    Generates a color wheel for optical flow visualization as presented in:
    Baker et al. "A Database and Evaluation Methodology for Optical Flow" (ICCV, 2007)
    URL: http://vision.middlebury.edu/flow/flowEval-iccv07.pdf.
    Also check the flow field color coding in the paper.

    Returns:
        colorwheel
    """
    RY = 15
    YG = 6
    GC = 4
    CB = 11
    BM = 13
    MR = 6

    ncols = RY + YG + GC + CB + BM + MR
    colorwheel = np.zeros([ncols, 3])
    col = 0

    # RY
    colorwheel[0:RY, 0] = 255
    colorwheel[0:RY, 1] = np.transpose(np.floor(255 * np.arange(0, RY) / RY))
    col += RY
    # YG
    colorwheel[col : col + YG, 0] = 255 - np.transpose(np.floor(255 * np.arange(0, YG) / YG))
    colorwheel[col : col + YG, 1] = 255
    col += YG
    # GC
    colorwheel[col : col + GC, 1] = 255
    colorwheel[col : col + GC, 2] = np.transpose(np.floor(255 * np.arange(0, GC) / GC))
    col += GC
    # CB
    colorwheel[col : col + CB, 1] = 255 - np.transpose(np.floor(255 * np.arange(0, CB) / CB))
    colorwheel[col : col + CB, 2] = 255
    col += CB
    # BM
    colorwheel[col : col + BM, 2] = 255
    colorwheel[col : col + BM, 0] = np.transpose(np.floor(255 * np.arange(0, BM) / BM))
    col += +BM
    # MR
    colorwheel[col : col + MR, 2] = 255 - np.transpose(np.floor(255 * np.arange(0, MR) / MR))
    colorwheel[col : col + MR, 0] = 255
    return colorwheel


def _normalized_flow_to_image(normalized_flow: np.ndarray) -> np.ndarray:
    H, W, _ = normalized_flow.shape

    colorwheel = _make_color_wheel()
    num_cols = colorwheel.shape[0]

    norm = np.sqrt(np.sum(normalized_flow**2, axis=-1))
    a = np.arctan2(-normalized_flow[..., 1], -normalized_flow[..., 0]) / np.pi
    fk = (a + 1) / 2 * (num_cols - 1)
    k0 = np.floor(fk).astype(np.int32)
    k1 = k0 + 1
    k1[k1 == num_cols] = 0
    f = fk - k0

    flow_image = np.zeros([H, W, 3], dtype=np.uint8)
    for c in range(colorwheel.shape[1]):
        tmp = colorwheel[..., c]
        col0 = tmp[k0] / 255.0
        col1 = tmp[k1] / 255.0
        col = (1 - f) * col0 + f * col1
        col = 1 - norm * (1 - col)
        flow_image[..., c] = np.floor(255.0 * col).astype(np.uint8)
    return flow_image


def flow_to_image(flow: np.ndarray) -> np.ndarray:
    """
    Convert flow to color image.

    Args:
        flow: optical flow of shape [H, W, 2]

    Returns:
        color image of shape [H, W, 3]
    """
    max_norm = np.sqrt(np.sum(flow**2, axis=-1)).max()
    eps = np.finfo(flow.dtype).eps
    normalized_flow = flow / (max_norm + eps)

    return _normalized_flow_to_image(normalized_flow)
